﻿using System;
using System.Collections.Generic;
using FunctionalCLR.Core;
using FunctionalCLR.Shell;
using Zove.Runtime;
using Zove.Runtime.FunctionalCLR;

namespace Zove.Shell.Core
{
    public class ZoveShellContext : ShellContext, IZoveShellContext
    {

        private DefaultFunctionManager fm = null;

        private ZoveShellParserFactory pf = null;

        public ZoveShellContext()
            : base()
        {
            this.fm = new ContextFunctionManager(this);
            this.pf = new ZoveShellParserFactory(this);
        }

        /// <summary>
        /// Run the script
        /// </summary>
        /// <param name="fileName"></param>
        public override void Run(string fileName)
        {
            ZoveShellInterpreter interpreter
                = new ZoveShellInterpreter(this);
            interpreter.Run(fileName);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="expression"></param>
        /// <returns></returns>
        public override object Evaluate(string expression)
        {
            Parser pa
                = this.GetParserFactory().CreateParser(
                expression, this);
            Node root = pa.Parse();
            if (root is ContextFuncNode)
            {
                (root as ContextFuncNode).Context = this;
            }
            object result = root.Evaluate();

            return result;
        }

        /// <summary>
        /// Add a new varible to the context
        /// </summary>
        /// <param name="factorName">the variable name</param>
        /// <returns></returns>
        //public override Node AddVariable(string factorName) {
        //    if (!IsLegalNaming(factorName)) {
        //        throw new IllegalNamingException(factorName);
        //    }

        //    RuntimeFactor rf = new RuntimeFactor(factorName);
        //    rf.Status = RuntimeFactorStatus.Unknown;

        //    FactorNode fn = new FactorNode(rf);

        //    return fn;
        //}

        /// <summary>
        /// Create a node expression
        /// </summary>
        /// <param name="expr"></param>
        /// <returns></returns>
        public override Node Parse(string expr)
        {
            Parser p = this.pf.CreateParser(expr, this);
            Node root = p.Parse();
            return root;
        }

        /// <summary>
        /// Get variable
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public override object GetVariable(string value)
        {
            int dotIndex = value.IndexOf('.');

            if (dotIndex < 0)
            {
                return base.GetVariable(value);

                //if (obj is FactorNode) {
                //    return obj as FactorNode;
                //} else if (obj is NumberNode) {
                //    return obj as NumberNode;
                //} else {
                //    return obj;
                //}
            }
            
            RuntimeFactor factor = GetFactor(value);
            if (factor != null)
            {
                return factor.Value;
            }

            return null;
            //if (instances.ContainsKey(instanceName))
            //{
            //    Solution solution = instances[instanceName];
            //    RuntimeModel rm = solution.Model;
            //    RuntimeFactor factor = rm.GetFactorByName(factorName);
            //    FactorNode fn = new FactorNode(factor);

            //    return fn;
            //}
            //else
            //{
            //    throw new Exception(
            //        string.Format(
            //        "Solution '{0}' is not instantiated.", instanceName));
            //}
        }

        /// <summary>
        /// Evaluate each instance
        /// </summary>
        public void EvaluateSolution(string variableName)
        {
            Instance instance = this.GetInstance(variableName);
            if (instance == null)
            {
                return;
            }

            instance.Evaluate();
        }

        protected Dictionary<string, Instance> instances
            = new Dictionary<string, Instance>();

        /// <summary>
        /// Add a new solution to the context
        /// </summary>
        /// <param name="name">name of the instance</param>
        /// <param name="instance">the solution instance</param>
        public void AddInstance(string name, Instance instance)
        {
            if (!IsLegalNaming(name))
            {
                throw new IllegalNamingException(name);
            }

            //SolutionNode sn = new SolutionNode(instance);
            StackFrame frame = base.Stack.Peek();
            if (frame.ContainsKey(name))
            {
                frame[name] = instance;
            }
            else
            {
                frame.Add(name, instance);
            }
        }

        public static bool IsLegalNaming(string name)
        {
            if (string.IsNullOrEmpty(name))
            {
                return false;
            }

            if (char.IsDigit(name[0]))
            {
                return false;
            }

            foreach (char c in name)
            {
                if (!(char.IsLetterOrDigit(c) || c == '_'))
                {
                    return false;
                }
            }

            return true;
        }

        /// <summary>
        /// Assign to a factor
        /// </summary>
        /// <param name="obj"></param>
        /// <param name="result"></param>
        /// <returns></returns>
        public override object Assign(
            string variableName, object result)
        {
            if (variableName.IndexOf('.') > 0)
            {
                RuntimeFactor factor = GetFactor(variableName);
                if (factor == null)
                {
                    throw new FactorNotDeclaredException(
                        variableName);
                }

                double value = Convert.ToDouble(result);
                factor.Value = value;
                factor.Status = RuntimeFactorStatus.Input;

                string instanceName = variableName.Split('.')[0];
                this.EvaluateSolution(instanceName);

                return result;
            }

            return base.Assign(variableName, result);
            //if (Utils.IsNumeric(result)) {
            //    FactorNode fn 
            //        = this.GetVariable(variableName) as FactorNode;
            //    if (fn == null) {
            //        if (!IsLegalNaming(variableName)) {
            //            throw new Exception(
            //                "The naming is not complied with the convention: " + variableName);
            //        }

            //        RuntimeFactor factor = new RuntimeFactor();
            //        factor.Name = variableName;
            //        factor.Value = Convert.ToDouble(result);

            //        fn = new FactorNode(factor);
            //        base.AddVariable(variableName, fn);
            //    }

            //    if (fn.Factor.Status == RuntimeFactorStatus.Unknown) {
            //        fn.Factor.Status = RuntimeFactorStatus.Input;
            //    } // if not, remain the current status

            //    fn.Factor.Value = Convert.ToDouble(result);

            //    this.EvaluateSolution();
            //    return fn;
            //}

            //return base.Assign(variableName, result);
        }

        /// <summary>
        /// Get a factor
        /// </summary>
        /// <param name="variableName"></param>
        /// <returns></returns>
        public RuntimeFactor GetFactor(string id)
        {
            if (id.IndexOf('.') < 0)
            {
                return null;
            }

            int index = id.IndexOf('.');
            string instanceName = id.Substring(0, index);
            string var = id.Substring(index+1, id.Length-index-1);
            Instance instance = this.GetInstance(instanceName);

            return instance.Model.GetFactorByName(var);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="id"></param>
        /// <returns></returns>
        public Instance GetInstance(string id)
        {
            StackFrame frame = this.Stack.Peek();
            if (frame.ContainsKey(id))
            {
                return frame[id] as Instance;
            }
            if (frame != RootFrame)
            {
                if(RootFrame.ContainsKey(id)) {
                    return RootFrame[id] as Instance;
                }
            }

            return null;

        }

        public object[] GetAllVariables()
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Remove the variable
        /// </summary>
        /// <param name="p">the variable to be removed</param>
        public void RemoveVariable(string p)
        {
            if (p.IndexOf('.') < 0)
            {
                base.RemoveVariable(p);
            }

            this.instances.Remove(p);
        }

        /// <summary>
        /// Check if the specific variable exists
        /// </summary>
        /// <param name="p"></param>
        /// <returns></returns>
        public override bool HasVariable(string p)
        {
            if (base.HasVariable(p))
            {
                return true;
            }

            return this.instances.ContainsKey(p);
        }

        public override IParserFactory GetParserFactory()
        {
            return pf;
        }

        public override IFunctionManager GetFunctionManager()
        {
            return this.fm;
        }

        public override IUnaryFunctionFactory UnaryFunctionFactory
        {
            get { return new ZoveShellUnaryFunctionFactory(this); }
        }
    }
}
