﻿using System.Text;
using EmbeddedSharp.Core;
using EmbeddedSharp.Core.Function.Arithmetics;
using EmbeddedSharp.Core.Function.Logic;
using EmbeddedSharp.Core.Function.Maths;
using EmbeddedSharp.Scripting.Function;

namespace Zove.Runtime.Core {
    public class UserDefinedFunctionConverter : IVisitor {
        protected ReplacingMode ReplacingMode;

        public UserDefinedFunctionConverter()
            : this(string.Empty, ReplacingMode.CompleteMatch) {
        }

        public UserDefinedFunctionConverter(string className)
            : this(className, ReplacingMode.CompleteMatch) {
        }

        public UserDefinedFunctionConverter(
            string className, ReplacingMode mode) {
            ClassName = className;
            ReplacingMode = mode;
        }

        public string ClassName { get; set; }

        #region IVisitor Members

        public object Visit(IdentifierNode node) {
            return node.Value;
        }

        public object Visit(NumberNode node) {
            return node.Value;
        }

        public object Visit(StringNode node) {
            return node.Value;
        }

        public object Visit(SubNode node) {
            object lr = node.Left.Accept(this);
            object rr = node.Right.Accept(this);

            return string.Format("({0} - {1})", lr, rr);
        }

        public object Visit(AddNode node) {
            object lr = node.Left.Accept(this);
            object rr = node.Right.Accept(this);

            return string.Format("({0} + {1})", lr, rr);
        }

        public object Visit(MulNode node) {
            object lr = node.Left.Accept(this);
            object rr = node.Right.Accept(this);

            return string.Format("({0} * {1})", lr, rr);
        }

        public object Visit(DivNode node) {
            object lr = node.Left.Accept(this);
            object rr = node.Right.Accept(this);

            return string.Format("({0} / {1})", lr, rr);
        }

        public object Visit(PowerNode node) {
            object lr = node.Left.Accept(this);
            object rr = node.Right.Accept(this);

            return string.Format("({0} ^ {1})", lr, rr);
        }

        public object Visit(LogNode node) {
            object lr = node.Left.Accept(this);
            object rr = node.Right.Accept(this);

            return string.Format("({0} ' {1})", lr, rr);
        }

        public object Visit(LargerThanNode node) {
            object lr = node.Left.Accept(this);
            object rr = node.Right.Accept(this);

            return string.Format("({0} > {1})", lr, rr);
        }

        public object Visit(LargerEqualToNode node) {
            object lr = node.Left.Accept(this);
            object rr = node.Right.Accept(this);

            return string.Format("({0} >= {1})", lr, rr);
        }

        public object Visit(LessThanNode node) {
            object lr = node.Left.Accept(this);
            object rr = node.Right.Accept(this);

            return string.Format("({0} < {1})", lr, rr);
        }

        public object Visit(LessEqualToNode node) {
            object lr = node.Left.Accept(this);
            object rr = node.Right.Accept(this);

            return string.Format("({0} <= {1})", lr, rr);
        }

        public object Visit(EqualNode node) {
            object lr = node.Left.Accept(this);
            object rr = node.Right.Accept(this);

            return string.Format("({0} == {1})", lr, rr);
        }

        public object Visit(NotEqualNode node) {
            object lr = node.Left.Accept(this);
            object rr = node.Right.Accept(this);

            return string.Format("({0} != {1})", lr, rr);
        }

        public object Visit(NotNode node) {
            object lr = node.Operand.Accept(this);

            return string.Format("(!{0})", lr);
        }

        public object Visit(FuncNode node) {
            if (node is UserDefinedRuleNode) {
                var udf
                    = node as UserDefinedRuleNode;
                string expr = udf.FuncBody;

                for (int index = udf.Arguments.Length - 1;
                     index >= 0;
                     index--) {
                    object value
                        = node.Parameters[index].Accept(this);
                    expr = Utils.ReplaceString(
                        udf.Arguments[index],
                        value.ToString(), expr,
                        ReplacingMode.StartingMatch
                        );
                }

                return expr;
            }
            if (node is UserDefinedFunctionNode) {
                var udf
                    = node as UserDefinedFunctionNode;
                string expr = udf.FuncBody;

                for (int index = udf.Arguments.Length - 1;
                     index >= 0;
                     index--) {
                    object value = node.Parameters[index].Accept(this);
                    expr = Utils.ReplaceString(
                        udf.Arguments[index],
                        value.ToString(), expr,
                        ReplacingMode.CompleteMatch
                        );
                }

                return expr;
            }

            var buff = new StringBuilder();

            buff.Append(node.Value).Append("(");
            for (int index = node.Parameters.Count - 1; index >= 0; index--) {
                object value = node.Parameters[index].Accept(this);
                buff.Append(value.ToString());
                if (index > 0) {
                    buff.Append(",");
                }
            }
            buff.Append(")");

            return buff.ToString();
        }

        public object Visit(AndNode node) {
            object lr = node.Left.Accept(this);
            object rr = node.Right.Accept(this);

            return string.Format("({0} && {1})", lr, rr);
        }

        public object Visit(OrNode node) {
            object lr = node.Left.Accept(this);
            object rr = node.Right.Accept(this);

            return string.Format("({0} || {1})", lr, rr);
        }

        public object Visit(ModNode node) {
            object lr = node.Left.Accept(this);
            object rr = node.Right.Accept(this);

            return string.Format("({0} % {1})", lr, rr);
        }

        #endregion

        public string Convert(Node root) {
            return (string)root.Accept(this);
        }
    }
}